@@ -8,14 +8,27 @@ __author__: Conor Heins, Alexander Tschantz, Brennan Klein
 
 import numpy as np
 import pandas as pd
+import seaborn as sns
+import matplotlib.pyplot as plt
 
 import warnings
 import itertools
 
+EPS_VAL = 1e-16 # global constant for use in norm_dist()
+
 def sample(probabilities):
-    sample_onehot = np.random.multinomial(1, probabilities.squeeze())
+    probabilities = probabilities.squeeze() if len(probabilities) > 1 else probabilities
+    sample_onehot = np.random.multinomial(1, probabilities)
     return np.where(sample_onehot == 1)[0][0]
 
+def sample_obj_array(arr):
+    """ 
+    Sample from set of Categorical distributions, stored in the sub-arrays of an object array 
+    """
+    
+    samples = [sample(arr_i) for arr_i in arr]
+
+    return samples
 
 def obj_array(num_arr):
     """
@@ -33,6 +46,24 @@ def obj_array_zeros(shape_list):
         arr[i] = np.zeros(shape)
     return arr
 
+def initialize_empty_A(num_obs, num_states):
+    """ 
+    Initializes an empty observation likelihood array or `A` array using a list of observation-modality dimensions (`num_obs`)
+    and hidden state factor dimensions (`num_states`)
+    """
+
+    A_shape_list = [ [no] + num_states for no in num_obs]
+    return obj_array_zeros(A_shape_list)
+
+def initialize_empty_B(num_states, num_controls):
+    """ 
+    Initializes an empty (controllable) transition likelihood array or `B` array using a list of hidden state factor dimensions (`num_states`)
+    and control factor dimensions (`num_controls)
+    """
+
+    B_shape_list = [ [ns, ns, num_controls[f]] for f, ns in enumerate(num_states)]
+    return obj_array_zeros(B_shape_list)
+
 def obj_array_uniform(shape_list):
     """ 
     Creates a numpy object array whose sub-arrays are uniform Categorical
@@ -121,12 +152,12 @@ def dirichlet_like(template_categorical, scale = 1.0):
     Helper function to construct a Dirichlet distribution based on an existing Categorical distribution
     """ 
 
-    if not is_arr_of_arr(template_categorical):
+    if not is_obj_array(template_categorical):
         warnings.warn(
                     "Input array is not an object array...\
                     Casting the input to an object array"
                 )
-        template_categorical = to_arr_of_arr(template_categorical)
+        template_categorical = to_obj_array(template_categorical)
 
     n_sub_arrays = len(template_categorical)
 
@@ -145,17 +176,17 @@ def get_model_dimensions(A=None, B=None):
                 )
 
     if A is not None:
-        num_obs = [a.shape[0] for a in A] if is_arr_of_arr(A) else [A.shape[0]]
+        num_obs = [a.shape[0] for a in A] if is_obj_array(A) else [A.shape[0]]
         num_modalities = len(num_obs)
     else:
         num_obs, num_modalities = None, None
     
     if B is not None:
-        num_states = [b.shape[0] for b in B] if is_arr_of_arr(B) else [B.shape[0]]
+        num_states = [b.shape[0] for b in B] if is_obj_array(B) else [B.shape[0]]
         num_factors = len(num_states)
     else:
         if A is not None:
-            num_states = list(A[0].shape[1:]) if is_arr_of_arr(A) else list(A.shape[1:])
+            num_states = list(A[0].shape[1:]) if is_obj_array(A) else list(A.shape[1:])
             num_factors = len(num_states)
         else:
             num_states, num_factors = None, None
@@ -201,17 +232,39 @@ def norm_dist_obj_arr(obj_arr):
     
     return normed_obj_array
 
+def is_normalized(dist):
+    """
+    Utility function for checking whether a single distribution or set of conditional categorical distributions is normalized.
+    Returns True if all distributions integrate to 1.0
+    """
 
-def is_arr_of_arr(arr):
+    if is_obj_array(dist):
+        normed_arrays = []
+        for i, arr in enumerate(dist):
+            column_sums = arr.sum(axis=0)
+            normed_arrays.append(np.allclose(column_sums, np.ones_like(column_sums)))
+        out = all(normed_arrays)
+    else:
+        column_sums = dist.sum(axis=0)
+        out = np.allclose(column_sums, np.ones_like(column_sums))
+    
+    return out
+
+def is_obj_array(arr):
     return arr.dtype == "object"
 
-
-def to_arr_of_arr(arr):
-    if is_arr_of_arr(arr):
+def to_obj_array(arr):
+    if is_obj_array(arr):
         return arr
-    arr_of_arr = obj_array(1)
-    arr_of_arr[0] = arr.squeeze()
-    return arr_of_arr
+    obj_array_out = obj_array(1)
+    obj_array_out[0] = arr.squeeze()
+    return obj_array_out
+
+def obj_array_from_list(list_input):
+    """
+    Takes a list of `numpy.ndarray` and converts them to a `numpy.ndarray` of `dtype = object`
+    """
+    return np.array(list_input, dtype = object)
 
 def process_observation_seq(obs_seq, n_modalities, n_observations):
     """
@@ -230,11 +283,23 @@ def process_observation_seq(obs_seq, n_modalities, n_observations):
 def process_observation(obs, num_modalities, num_observations):
     """
     Helper function for formatting observations    
-
-        Observations can either be `int` (converted to one-hot)
-        `tuple` (obs for each modality), or `list` (obs for each modality)
+    USAGE NOTES:
+    - If `obs` is a 1D numpy array, it must be a one-hot vector, where one entry (the entry of the observation) is 1.0 
+    and all other entries are 0. This therefore assumes it's a single modality observation. If these conditions are met, then
+    this function will return `obs` unchanged. Otherwise, it'll throw an error.
+    - If `obs` is an int, it assumes this is a single modality observation, whose observation index is given by the value of `obs`. This function will convert
+    it to be a one hot vector.
+    - If `obs` is a list, it assumes this is a multiple modality observation, whose len is equal to the number of observation modalities,
+    and where each entry `obs[m]` is the index of the observation, for that modality. This function will convert it into an object array
+    of one-hot vectors.
+    - If `obs` is a tuple, same logic as applies for list (see above).
+    - if `obs` is a numpy object array (array of arrays), this function will return `obs` unchanged.
     """
 
+    if isinstance(obs, np.ndarray) and not is_obj_array(obs):
+        assert num_modalities == 1, "If `obs` is a 1D numpy array, `num_modalities` must be equal to 1"
+        assert len(np.where(obs)[0]) == 1, "If `obs` is a 1D numpy array, it must be a one hot vector (e.g. np.array([0.0, 1.0, 0.0, ....]))"
+
     if isinstance(obs, (int, np.integer)):
         obs = onehot(obs, num_observations[0])
 
@@ -512,34 +577,34 @@ def convert_B_stubs_to_ndarray(B_stubs, model_labels):
 
     return B
 
-def build_belief_array(qx):
+# def build_belief_array(qx):
 
-    """
-    This function constructs array-ified (not nested) versions
-    of the posterior belief arrays, that are separated 
-    by policy, timepoint, and hidden state factor
-    """
+#     """
+#     This function constructs array-ified (not nested) versions
+#     of the posterior belief arrays, that are separated 
+#     by policy, timepoint, and hidden state factor
+#     """
 
-    num_policies = len(qx)
-    num_timesteps = len(qx[0])
-    num_factors = len(qx[0][0])
+#     num_policies = len(qx)
+#     num_timesteps = len(qx[0])
+#     num_factors = len(qx[0][0])
 
-    if num_factors > 1:
-        belief_array = utils.obj_array(num_factors)
-        for factor in range(num_factors):
-            belief_array[factor] = np.zeros( (num_policies, qx[0][0][factor].shape[0], num_timesteps) )
-        for policy_i in range(num_policies):
-            for timestep in range(num_timesteps):
-                for factor in range(num_factors):
-                    belief_array[factor][policy_i, :, timestep] = qx[policy_i][timestep][factor]
-    else:
-        num_states = qx[0][0][0].shape[0]
-        belief_array = np.zeros( (num_policies, num_states, num_timesteps) )
-        for policy_i in range(num_policies):
-            for timestep in range(num_timesteps):
-                belief_array[policy_i, :, timestep] = qx[policy_i][timestep][0]
+#     if num_factors > 1:
+#         belief_array = obj_array(num_factors)
+#         for factor in range(num_factors):
+#             belief_array[factor] = np.zeros( (num_policies, qx[0][0][factor].shape[0], num_timesteps) )
+#         for policy_i in range(num_policies):
+#             for timestep in range(num_timesteps):
+#                 for factor in range(num_factors):
+#                     belief_array[factor][policy_i, :, timestep] = qx[policy_i][timestep][factor]
+#     else:
+#         num_states = qx[0][0][0].shape[0]
+#         belief_array = np.zeros( (num_policies, num_states, num_timesteps) )
+#         for policy_i in range(num_policies):
+#             for timestep in range(num_timesteps):
+#                 belief_array[policy_i, :, timestep] = qx[policy_i][timestep][0]
     
-    return belief_array
+#     return belief_array
 
 def build_xn_vn_array(xn):
 
@@ -554,9 +619,9 @@ def build_xn_vn_array(xn):
     num_factors = len(xn[0][0])
 
     if num_factors > 1:
-        xn_array = utils.obj_array(num_factors)
+        xn_array = obj_array(num_factors)
         for factor in range(num_factors):
-            num_states, infer_len = xn[0][0][f].shape
+            num_states, infer_len = xn[0][0][factor].shape
             xn_array[factor] = np.zeros( (num_itr, num_states, infer_len, num_policies) )
         for policy_i in range(num_policies):
             for itr in range(num_itr):
@@ -571,6 +636,31 @@ def build_xn_vn_array(xn):
     
     return xn_array
 
+def plot_beliefs(belief_dist, title=""):
+    """
+    Utility function that plots a bar chart of a categorical probability distribution,
+    with each bar height corresponding to the probability of one of the elements of the categorical
+    probability vector.
+    """
+
+    plt.grid(zorder=0)
+    plt.bar(range(belief_dist.shape[0]), belief_dist, color='r', zorder=3)
+    plt.xticks(range(belief_dist.shape[0]))
+    plt.title(title)
+    plt.show()
+    
+def plot_likelihood(A, title=""):
+    """
+    Utility function that shows a heatmap of a 2-D likelihood (hidden causes in the columns, observations in the rows),
+    with hotter colors indicating higher probability.
+    """
+
+    ax = sns.heatmap(A, cmap="OrRd", linewidth=2.5)
+    plt.xticks(range(A.shape[1]+1))
+    plt.yticks(range(A.shape[0]+1))
+    plt.title(title)
+    plt.show()
+
 
     
    
